# ----------------------------------------------------------------------------
#
#  Welcome to Baml! To use this generated code, please run the following:
#
#  $ pip install baml
#
# ----------------------------------------------------------------------------

# This file was generated by BAML: please do not edit it. Instead, edit the
# BAML files and re-generate this code using: baml-cli generate
# baml-cli is available with the baml package.

import typing
import typing_extensions
import baml_py

from . import stream_types, types, type_builder
from .parser import LlmResponseParser, LlmStreamParser
from .runtime import DoNotUseDirectlyCallManager, BamlCallOptions
from .globals import DO_NOT_USE_DIRECTLY_UNLESS_YOU_KNOW_WHAT_YOURE_DOING_RUNTIME as __runtime__


class BamlAsyncClient:
    __options: DoNotUseDirectlyCallManager
    __stream_client: "BamlStreamClient"
    __http_request: "BamlHttpRequestClient"
    __http_stream_request: "BamlHttpStreamRequestClient"
    __llm_response_parser: LlmResponseParser
    __llm_stream_parser: LlmStreamParser

    def __init__(self, options: DoNotUseDirectlyCallManager):
        self.__options = options
        self.__stream_client = BamlStreamClient(options)
        self.__http_request = BamlHttpRequestClient(options)
        self.__http_stream_request = BamlHttpStreamRequestClient(options)
        self.__llm_response_parser = LlmResponseParser(options)
        self.__llm_stream_parser = LlmStreamParser(options)

    def with_options(self,
        tb: typing.Optional[type_builder.TypeBuilder] = None,
        client_registry: typing.Optional[baml_py.baml_py.ClientRegistry] = None,
        collector: typing.Optional[typing.Union[baml_py.baml_py.Collector, typing.List[baml_py.baml_py.Collector]]] = None,
        env: typing.Optional[typing.Dict[str, typing.Optional[str]]] = None,
    ) -> "BamlAsyncClient":
        options: BamlCallOptions = {}
        if tb is not None:
            options["tb"] = tb
        if client_registry is not None:
            options["client_registry"] = client_registry
        if collector is not None:
            options["collector"] = collector
        if env is not None:
            options["env"] = env
        return BamlAsyncClient(self.__options.merge_options(options))

    @property
    def stream(self):
      return self.__stream_client

    @property
    def request(self):
      return self.__http_request

    @property
    def stream_request(self):
      return self.__http_stream_request

    @property
    def parse(self):
      return self.__llm_response_parser

    @property
    def parse_stream(self):
      return self.__llm_stream_parser
    
    async def RAG(self, question: str,context: str,
        baml_options: BamlCallOptions = {},
    ) -> types.Response:
        result = await self.__options.merge_options(baml_options).call_function_async(function_name="RAG", args={
            "question": question,"context": context,
        })
        return typing.cast(types.Response, result.cast_to(types, types, stream_types, False, __runtime__))
    


class BamlStreamClient:
    __options: DoNotUseDirectlyCallManager

    def __init__(self, options: DoNotUseDirectlyCallManager):
        self.__options = options

    def RAG(self, question: str,context: str,
        baml_options: BamlCallOptions = {},
    ) -> baml_py.BamlStream[stream_types.Response, types.Response]:
        ctx, result = self.__options.merge_options(baml_options).create_async_stream(function_name="RAG", args={
            "question": question,"context": context,
        })
        return baml_py.BamlStream[stream_types.Response, types.Response](
          result,
          lambda x: typing.cast(stream_types.Response, x.cast_to(types, types, stream_types, True, __runtime__)),
          lambda x: typing.cast(types.Response, x.cast_to(types, types, stream_types, False, __runtime__)),
          ctx,
        )
    

class BamlHttpRequestClient:
    __options: DoNotUseDirectlyCallManager

    def __init__(self, options: DoNotUseDirectlyCallManager):
        self.__options = options

    async def RAG(self, question: str,context: str,
        baml_options: BamlCallOptions = {},
    ) -> baml_py.baml_py.HTTPRequest:
        result = await self.__options.merge_options(baml_options).create_http_request_async(function_name="RAG", args={
            "question": question,"context": context,
        }, mode="request")
        return result
    

class BamlHttpStreamRequestClient:
    __options: DoNotUseDirectlyCallManager

    def __init__(self, options: DoNotUseDirectlyCallManager):
        self.__options = options

    async def RAG(self, question: str,context: str,
        baml_options: BamlCallOptions = {},
    ) -> baml_py.baml_py.HTTPRequest:
        result = await self.__options.merge_options(baml_options).create_http_request_async(function_name="RAG", args={
            "question": question,"context": context,
        }, mode="stream")
        return result
    

b = BamlAsyncClient(DoNotUseDirectlyCallManager({}))


# ----------------------------------------------------------------------------
#
#  Welcome to Baml! To use this generated code, please run the following:
#
#  $ pip install baml
#
# ----------------------------------------------------------------------------

# This file was generated by BAML: please do not edit it. Instead, edit the
# BAML files and re-generate this code using: baml-cli generate
# baml-cli is available with the baml package.

_file_map = {

    "clients.baml": "// Learn more about clients at https://docs.boundaryml.com/docs/snippets/clients/overview\n\nclient<llm> CustomGPT4o {\n  provider openai\n  options {\n    model \"gpt-4o\"\n    api_key env.OPENAI_API_KEY\n  }\n}\n\nclient<llm> CustomGPT4oMini {\n  provider openai\n  retry_policy Exponential\n  options {\n    model \"gpt-4o-mini\"\n    api_key env.OPENAI_API_KEY\n  }\n}\n\nclient<llm> CustomSonnet {\n  provider anthropic\n  options {\n    model \"claude-3-5-sonnet-20241022\"\n    api_key env.ANTHROPIC_API_KEY\n  }\n}\n\n\nclient<llm> CustomHaiku {\n  provider anthropic\n  retry_policy Constant\n  options {\n    model \"claude-3-haiku-20240307\"\n    api_key env.ANTHROPIC_API_KEY\n  }\n}\n\n// https://docs.boundaryml.com/docs/snippets/clients/round-robin\nclient<llm> CustomFast {\n  provider round-robin\n  options {\n    // This will alternate between the two clients\n    strategy [CustomGPT4oMini, CustomHaiku]\n  }\n}\n\n// https://docs.boundaryml.com/docs/snippets/clients/fallback\nclient<llm> OpenaiFallback {\n  provider fallback\n  options {\n    // This will try the clients in order until one succeeds\n    strategy [CustomGPT4oMini, CustomGPT4oMini]\n  }\n}\n\n// https://docs.boundaryml.com/docs/snippets/clients/retry\nretry_policy Constant {\n  max_retries 3\n  // Strategy is optional\n  strategy {\n    type constant_delay\n    delay_ms 200\n  }\n}\n\nretry_policy Exponential {\n  max_retries 2\n  // Strategy is optional\n  strategy {\n    type exponential_backoff\n    delay_ms 300\n    multiplier 1.5\n    max_delay_ms 10000\n  }\n}",
    "generators.baml": "// This helps use auto generate libraries you can use in the language of\n// your choice. You can have multiple generators if you use multiple languages.\n// Just ensure that the output_dir is different for each generator.\ngenerator target {\n    // Valid values: \"python/pydantic\", \"typescript\", \"ruby/sorbet\", \"rest/openapi\"\n    output_type \"python/pydantic\"\n\n    // Where the generated code will be saved (relative to baml_src/)\n    output_dir \"../\"\n\n    // The version of the BAML package you have installed (e.g. same version as your baml-py or @boundaryml/baml).\n    // The BAML VSCode extension version should also match this version.\n    version \"0.203.1\"\n\n    // Valid values: \"sync\", \"async\"\n    // This controls what `b.FunctionName()` will be (sync or async).\n    default_client_mode sync\n}\n",
    "rag.baml": "// --- Data Models ---\n\nclass Response {\n  question_type \"verbatim\" | \"reworded\" | \"new\"\n  question string\n  answer string | null @description(\"A concise answer to the question in 1-3 sentences\")\n}\n\nclient<llm> GenericClient {\n  provider \"openai-generic\"\n  options {\n    model gemma3:12b\n    base_url \"http://localhost:11434/v1\"\n    temperature 0.0\n  }\n}\n\n\n// --- Functions ---\n\nfunction RAG(question: string, context: string) -> Response {\n  client GenericClient\n  prompt #\"\n    Answer the question using the provided context.\n    Do not make up an answer. If you don't know the answer, return a null response.\n    Always strive to provide concise, helpful, and context-aware answers.\n     \n    CONTEXT: {{ context }}\n\n    {{ ctx.output_format }}\n\n    QUESTION: {{ question }}\n\n    RESPONSE:\n  \"#\n}\n\n\n// --- Tests ---\n\ntest TestMissingContext{\n  functions [RAG]\n  args {\n    question \"Who founded Lantana Trade LLP?\"\n    context #\"\n      BoundaryML is the company that makes BAML, the best way to get structured with LLMs.\n    \"#\n  }\n}\n",
}

def get_baml_files():
    return _file_map


#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Construct the knowledge graph.
see copyright/license https://github.com/DerwenAI/strwythura/README.md
"""

from collections import defaultdict
import typing

from icecream import ic  # type: ignore
import lancedb  # type: ignore
import networkx as nx
import polars as pl
import spacy

from .graph import Entity, TextChunk
from .nlp import Parser
from .scrape import Scraper
from .textrank import run_textrank, cooccur_entities


class KnowledgeGraph:
    """
Construct a _knowledge graph_ and build out assets to serialize then
use later.
    """

    def __init__ (
        self,
        config: dict,
        ) -> None:
        """
Constructor.
        """
        self.config: dict = config
        self.known_lemma: typing.List[ str ] = []


    def build_graph (  # pylint: disable=R0912,R0913,R0914,R0917,W0102
        self,
        parser: Parser,
        simple_pipe: spacy.Language,
        entity_pipe: spacy.Language,
        chunk_table: lancedb.table.LanceTable,
        sem_overlay: nx.Graph,
        w2v_vectors: list = [],
        *,
        debug: bool = False,
        ) -> None:
        """
Construct a knowledge graph from unstructured data sources.
        """
        # iterate through the URL list, scraping text and building chunks
        scraper: Scraper = Scraper(self.config, parser)
        chunk_id: int = 0

        for url in parser.url_list:
            # define data structures intialized for each parsed document
            lex_graph: nx.Graph = nx.Graph()
            chunk_list: typing.List[ TextChunk ] = []

            chunk_id = scraper.scrape_html(
                simple_pipe,
                url,
                chunk_list,
                chunk_id,
            )

            chunk_table.add(chunk_list)

            # parse each chunk to build a lexical graph per source URL
            for chunk in chunk_list:
                span_decoder: typing.Dict[ tuple, Entity ] = {}

                doc: spacy.tokens.doc.Doc = parser.parse_text(  # pylint: disable=I1101
                    entity_pipe,
                    self.known_lemma,
                    lex_graph,
                    chunk,
                    debug = debug,
                )

                if debug:
                    ic(chunk)

                # keep track of sentence numbers per chunk, to use later
                # for entity co-occurrence links
                sent_map: typing.Dict[ spacy.tokens.span.Span, int ] = {}  # pylint: disable=I1101

                for sent_id, sent in enumerate(doc.sents):
                    sent_map[sent] = sent_id

                # classify the recognized spans within this chunk as
                # potential entities

                # NB: if we'd run [_entity resolution_]
                # see: <https://neo4j.com/developer-blog/entity-resolved-knowledge-graphs/>
                # previously from _structured_ or _semi-structured_ data sources to
                # generate a "backbone" for the knowledge graph, then we could use
                # contextualized _surface forms_ perform _entity linking_ on the
                # entities extracted here from _unstructured_ data

                for span in doc.ents:
                    self.make_entity(
                        span_decoder,
                        sent_map,
                        span,
                        chunk,
                        debug = debug,
                    )

                for span in doc.noun_chunks:
                    self.make_entity(
                        span_decoder,
                        sent_map,
                        span,
                        chunk,
                        debug = False, # debug
                    )

                # overlay the recognized entity spans atop the base layer
                # constructed by _textgraph_ analysis of the `spaCy` parse trees
                for ent in span_decoder.values():
                    if ent.key not in parser.STOP_WORDS:
                        self.extract_entity(
                            lex_graph,
                            ent,
                            debug = debug,
                        )

                # extract relations for co-occurring entity pairs
                ## PLACEHOLDER

                # connect entities which co-occur within the same sentence
                cooccur_entities(
                    lex_graph,
                    span_decoder,
                )

                # build the vector input for entity embeddings
                w2v_map: typing.Dict[ int, typing.Set[ str ]] = defaultdict(set)

                for ent in span_decoder.values():
                    if ent.node is not None:
                        w2v_map[ent.sent_id].add(ent.key)

                for sent_id, ents in w2v_map.items():
                    vec: list = list(ents)
                    vec.insert(0, str(sent_id))
                    w2v_vectors.append(vec)

            # apply _textrank_ to the graph (in the url/doc iteration)
            # then report the top-ranked extracted entities
            df: pl.DataFrame = run_textrank(
                self.config,
                lex_graph,
            )

            if debug:
                ic(url, df.head(11))

            # abstract a semantic overlay from the lexical graph
            # and persist this in the resulting KG
            self.abstract_overlay(
                url,
                chunk_list,
                lex_graph,
                sem_overlay,
            )

            if debug:
                print("nodes", len(sem_overlay.nodes), "edges", len(sem_overlay.edges))


    def abstract_overlay (  # pylint: disable=R0914
        self,
        url: str,
        chunk_list: typing.List[ TextChunk ],
        lex_graph: nx.Graph,
        sem_overlay: nx.Graph,
        ) -> None:
        """
Abstract a _semantic overlay_ from the lexical graph -- in other words
which nodes and edges get promoted up to the next level?

Also connect the extracted entities with their source chunks, where
the latter first-class citizens within the KG.
        """
        kept_nodes: typing.Set[ int ] = set()
        skipped_rel: typing.Set[ str ] = set([
            "FOLLOWS_LEXICALLY",
            "COMPOUND_ELEMENT_OF",
        ])

        chunk_nodes: typing.Dict[ int, str ] = {
            chunk.uid: f"chunk_{chunk.uid}"
            for chunk in chunk_list
        }

        for chunk_id, node_id in chunk_nodes.items():
            sem_overlay.add_node(
                node_id,
                kind = "Chunk",
                chunk = chunk_id,
                url = url,
            )

        for node_id, node_attr in lex_graph.nodes(data = True):
            if node_attr["kind"] == "Entity":
                kept_nodes.add(node_id)
                count: int = node_attr["count"]

                if not sem_overlay.has_node(node_id):
                    sem_overlay.add_node(
                        node_id,
                        kind = "Entity",
                        key = node_attr["key"],
                        text = node_attr["text"],
                        label = node_attr["label"],
                        rank = node_attr["rank"],
                        count = count,
                    )
                else:
                    sem_overlay.nodes[node_id]["count"] += count

                sem_overlay.add_edge(
                    node_id,
                    chunk_nodes[node_attr["chunk"]],
                    rel = "WITHIN",
                    weight = node_attr["rank"],
                )

        for src_id, dst_id, edge_attr in lex_graph.edges(data = True):
            if src_id in kept_nodes and dst_id in kept_nodes:
                rel: str = edge_attr["rel"]
                prob: float = 1.0

                if "prob" in edge_attr:
                    prob = edge_attr["prob"]

                if rel not in skipped_rel:
                    if not sem_overlay.has_edge(src_id, dst_id):
                        sem_overlay.add_edge(
                            src_id,
                            dst_id,
                            rel = rel,
                            prob = prob,
                        )
                    else:
                        sem_overlay[src_id][dst_id]["prob"] = max(
                            prob,
                            sem_overlay.edges[(src_id, dst_id)]["prob"],
                        )


    def make_entity (  # pylint: disable=R0913
        self,
        span_decoder: typing.Dict[ tuple, Entity ],
        sent_map: typing.Dict[ spacy.tokens.span.Span, int ],  # pylint: disable=I1101
        span: spacy.tokens.span.Span,  # pylint: disable=I1101
        chunk: TextChunk,
        *,
        debug: bool = False,  # pylint: disable=W0613
        ) -> Entity:
        """
Instantiate one `Entity` object, adding to our working "vocabulary".
        """
        key: str = " ".join([
            tok.pos_ + "." + tok.lemma_.strip().lower()
            for tok in span
        ])

        ent: Entity = Entity(
            ( span.start, span.end, ),
            key,
            span.text,
            span.label_,
            chunk.uid,
            sent_map[span.sent],
            span,
        )

        if ent.loc not in span_decoder:
            span_decoder[ent.loc] = ent

            if False: # debug  # pylint: disable=W0125
                ic(ent)

        return ent


    def extract_entity (
        self,
        lex_graph: nx.Graph,
        ent: Entity,
        *,
        debug: bool = False,  # pylint: disable=W0613
        ) -> None:
        """
Link one `Entity` into this doc's lexical graph.
        """
        prev_known: bool = False

        if ent.key not in self.known_lemma:
            # add a new Entity node to the graph and link to its component Lemma nodes
            self.known_lemma.append(ent.key)
        else:
            # phrase for this entity has been previously seen in other documents
            prev_known = True

        node_id: int = self.known_lemma.index(ent.key)
        ent.node = node_id

        # hydrate a compound phrase in this doc's lexical graph
        if not lex_graph.has_node(node_id):
            lex_graph.add_node(
                node_id,
                key = ent.key,
                kind = "Entity",
                label = ent.label,
                pos = "NP",
                text = ent.text,
                chunk = ent.chunk_id,
                count = 1,
            )

            for tok in ent.span:
                tok_key: str = tok.pos_ + "." + tok.lemma_.strip().lower()

                if tok_key in self.known_lemma:
                    tok_idx: int = self.known_lemma.index(tok_key)

                    lex_graph.add_edge(
                        node_id,
                        tok_idx,
                        rel = "COMPOUND_ELEMENT_OF",
                    )

        if prev_known:
            # promote a previous Lemma node to an Entity
            node: dict = lex_graph.nodes[node_id]
            node["kind"] = "Entity"
            node["chunk"] = ent.chunk_id
            node["count"] += 1

            # select the more specific label
            if "label" not in node or node["label"] == "NP":
                node["label"] = ent.label

        if False: # debug  # pylint: disable=W0125
            ic(ent)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Interactive visualization of a `NetworkX` graph based on `PyVis`
see copyright/license https://github.com/DerwenAI/strwythura/README.md
"""

import math

import networkx as nx
import pyvis  # type: ignore


def gen_pyvis (
    graph: nx.Graph,
    html_file: str,
    *,
    num_docs: int = 1,
    notebook: bool = False,
    ) -> None:
    """
Use `pyvis` to provide an interactive visualization of the graph layers.
    """
    pv_net: pyvis.network.Network = pyvis.network.Network(
        height = "900px",
        width = "100%",
        notebook = notebook,
        cdn_resources = "remote",
    )

    for node_id, node_attr in graph.nodes(data = True):
        if node_attr.get("kind") == "Entity":
            color: str = "hsla(65, 46%, 58%, 0.80)"
            size: int = round(20 * math.log(1.0 + math.sqrt(float(node_attr.get("count"))) / num_docs))  # type: ignore # pylint: disable=C0301
            label: str = node_attr.get("text")  # type: ignore
            title: str = node_attr.get("key")  # type: ignore
        else:
            color = "hsla(306, 45%, 57%, 0.95)"
            size = 5
            label = node_id
            title = node_attr.get("url")  # type: ignore

        pv_net.add_node(
            node_id,
            label = label,
            title = title,
            color = color,
            size = size,
        )

    for src_node, dst_node, edge_attr in graph.edges(data = True):
        pv_net.add_edge(
            src_node,
            dst_node,
            title = edge_attr.get("rel"),
        )

        pv_net.toggle_physics(True)
        pv_net.show_buttons(filter_ = [ "physics" ])
        pv_net.save_graph(html_file)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Builds assets for constructing a KG, then running GraphRAG downstream.
see copyright/license https://github.com/DerwenAI/strwythura/README.md
"""

import itertools
import json
import logging
import os
import pathlib
import tomllib
import traceback
import typing
import warnings

from icecream import ic  # type: ignore
from rdflib.namespace import RDF, SKOS
import gensim  # type: ignore
import lancedb  # type: ignore
import networkx as nx
import polars as pl
import rdflib
import spacy
import transformers

from .baml_client import b
from .baml_client import types as baml_types
from .graph import TextChunk
from .kg import KnowledgeGraph
from .nlp import Parser
from .vis import gen_pyvis


class Strwythura:  # pylint: disable=R0902
    """
Builds assets for constructing a KG, then running GraphRAG downstream.
    """

    def __init__ (
        self,
        *,
        config_path: pathlib.Path = pathlib.Path("config.toml"),
        ) -> None:
        """
Constructor.
        """
        # configuration
        self.config: dict = {}

        with open(config_path, mode = "rb") as fp:
            self.config = tomllib.load(fp)

        # disable noisy logging
        os.environ["BAML_LOG"] = "WARN"
        os.environ["TOKENIZERS_PARALLELISM"] = "0"

        logging.disable(logging.ERROR)
        transformers.logging.set_verbosity_error()

        ## none of this works!
        #os.environ["TQDM_DISABLE"] = "1"
        #loguru.logger.disable(gliner_spacy.pipeline.__name__)
        #loggers: dict =
        # { name:logging.getLogger(name) for name in logging.root.manager.loggerDict }
        #ic(loggers)

        # load the semantic layer: define a context for the domain
        domain_path: pathlib.Path = pathlib.Path(self.config["kg"]["domain_path"])
        self.domain_graph: rdflib.Graph = rdflib.Graph()

        self.domain_graph.parse(
            domain_path.as_posix(),
            format = "turtle",
        )

        # initialize the data structures used for assets
        self.parser: Parser = Parser(self.config)

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")

            self.simple_pipe: spacy.Language = spacy.load(self.config["nlp"]["spacy_model"])
            self.entity_pipe: typing.Optional[ spacy.Language ] = None
            self.chunk_table: typing.Optional[ lancedb.table.LanceTable ] = None
            self.sem_overlay: nx.Graph = nx.Graph()
            self.w2v_vectors: list = []
            self.w2v_model: typing.Optional[ gensim.models.Word2Vec ] = None


    def get_ner_labels (
        self,
        ) -> typing.List[ str ]:
        """
Extract the labels used for zero-shot NER and corresponding graph
nodes within the semantic layer definition for this domain context.
        """
        return [
            str(label)
            for concept in self.domain_graph.subjects(RDF.type, SKOS.Concept)
            for label in self.domain_graph.objects(concept, SKOS.prefLabel, unique = True)
        ]


    def build_assets (  # pylint: disable=R0913
        self,
        url_list: typing.List[ str ],
        *,
        debug: bool = False,
        kg_path: typing.Optional[ pathlib.Path ] = None,
        w2v_path: typing.Optional[ pathlib.Path ] = None,
        ) -> None:
        """
Builds assets for constructing a KG.
        """
        self.parser.update_data(
            url_list,
            self.get_ner_labels(),
        )

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")

            try:
                # given the NER labels, build the `spaCy` pipe
                self.entity_pipe = self.parser.build_entity_pipe()

                # initialize the chunk table
                vect_db: lancedb.db.LanceDBConnection = lancedb.connect(self.config["vect"]["lancedb_uri"])  # pylint: disable=C0301

                self.chunk_table = vect_db.create_table(
                    self.config["vect"]["chunk_table"],
                    schema = TextChunk,
                    mode = "overwrite",
                )

                # construct the graph
                kg: KnowledgeGraph = KnowledgeGraph(self.config)

                kg.build_graph(
                    self.parser,
                    self.simple_pipe,
                    self.entity_pipe,
                    self.chunk_table,
                    self.sem_overlay,
                    self.w2v_vectors,
                    debug = debug,
                )

                # serialize assets
                self.embed_entities(w2v_path = w2v_path)
                self.save_graph(kg_path = kg_path)

            except Exception as ex:  # pylint: disable=W0718
                ic(ex)
                traceback.print_exc()


    def embed_entities (
        self,
        *,
        w2v_path: typing.Optional[ pathlib.Path ] = None,
        ) -> None:
        """
Train a `gensim.Word2Vec` model for entity embeddings.
        """
        w2v_max: int = max([  # pylint: disable=R1728
            len(vec) - 1
            for vec in self.w2v_vectors
        ])

        self.w2v_model = gensim.models.Word2Vec(
            self.w2v_vectors,
            min_count = 2,
            window = w2v_max,
        )

        if w2v_path is None:
            w2v_path = pathlib.Path(self.config["ent"]["w2v_path"])

        self.w2v_model.save(w2v_path.as_posix())


    def save_graph (
        self,
        *,
        kg_path: typing.Optional[ pathlib.Path ] = None,
        ) -> None:
        """
Serialize the KG
        """
        if kg_path is None:
            kg_path = pathlib.Path(self.config["kg"]["kg_path"])

        with kg_path.open("w", encoding = "utf-8") as fp:
            fp.write(
                json.dumps(
                    nx.node_link_data(
                        self.sem_overlay,
                        edges = "edges",
                    ),
                    indent = 2,
                    sort_keys = True,
                )
            )


    def gen_visualization (
        self,
        *,
        html_path: typing.Optional[ pathlib.Path ] = None,
        ) -> None:
        """
Generate HTML for an interactive visualization of the graph, based on `PyVis`
        """
        if html_path is None:
            html_path = pathlib.Path(self.config["kg"]["html_path"])

        gen_pyvis(
            self.sem_overlay,
            html_path.as_posix(),
            num_docs = len(self.parser.url_list),
        )


    def load_assets (
        self,
        *,
        kg_path: typing.Optional[ pathlib.Path ] = None,
        w2v_path: typing.Optional[ pathlib.Path ] = None,
        ) -> None:
        """
Load the serialized assets for a constructed KG.
        """
        vect_db: lancedb.db.LanceDBConnection = lancedb.connect(self.config["vect"]["lancedb_uri"])
        self.chunk_table = vect_db.open_table(self.config["vect"]["chunk_table"])

        if w2v_path is None:
            w2v_path = pathlib.Path(self.config["ent"]["w2v_path"])

        self.w2v_model = gensim.models.Word2Vec.load(w2v_path.as_posix())

        if kg_path is None:
            kg_path = pathlib.Path(self.config["kg"]["kg_path"])

        with pathlib.Path(kg_path).open("r", encoding = "utf-8") as fp:
            self.sem_overlay = nx.node_link_graph(
                json.load(fp),
                edges = "edges",
            )

        # build the `spaCy` pipe, no need for input URL list
        self.parser.update_data(
            [],
            self.get_ner_labels(),
        )

        self.entity_pipe = self.parser.build_entity_pipe()


class GraphRAG:
    """
Run an example query through LanceDB to identify _chunks_ and through
the Word2Vec entity embedding model for a _semantic expansion_ to
produce a set of _anchor nodes_ in the NetworkX graph.
    """

    def __init__ (
        self,
        strw: Strwythura,
        ) -> None:
        """
Constructor.
        """
        self.strw: Strwythura = strw


    def find_entities (
        self,
        question: str,
        ) -> typing.Iterator[ str ]:
        """
Extract entity spans from a text question.
        """
        doc: spacy.tokens.doc.Doc = self.strw.entity_pipe(question)  # type: ignore  # pylint: disable=I1101

        for span in doc.ents:
            key: str = " ".join([
                tok.pos_ + "." + tok.lemma_.strip().lower()
                for tok in span
            ])

            yield key


    def get_chunks (  # pylint: disable=R0914
        self,
        question: str,
        *,
        debug: bool = False,
        num_chunks: typing.Optional[ int ] = None,
        ) -> typing.List[ str ]:
        """
Run semantic search to produce a set of text chunks.
        """
        if num_chunks is None:
            num_chunks = self.strw.config["rag"]["num_chunks"]

        # extract entities from the question
        entities: typing.Set[ str ] = set(list(self.find_entities(question)))

        if debug:
            ic(entities)

        # semantic expansion using entity embeddings
        neighbors: typing.Set[ str ] = set()

        try:
            for entity in entities:
                neighbor_iter = self.strw.w2v_model.wv.most_similar(  # type: ignore
                    positive = [ entity ],
                    topn = num_chunks,
                )

                for neighbor in neighbor_iter:
                    neighbors.add(neighbor)
        except KeyError:
            pass

        if debug:
            ic(neighbors)

        # map the expanded set of entities to nodes in the graph
        expanded_entities: set = entities.union(neighbors)

        anchor_nodes: set = {
            node
            for node, dat in self.strw.sem_overlay.nodes(data = True)
            if "key" in dat and dat["key"] in expanded_entities
        }

        if debug:
            ic(anchor_nodes)

        # extract a subgraph based on shortest paths between anchor nodes
        node_iter: typing.Iterator[ int ] = self.extract_subgraph(
            anchor_nodes,
            debug = debug,
        )

        # extract the chunk neighbors
        chunk_iter: typing.Iterator[ int ] = self.extract_chunk_neighbors(
            list(node_iter),
            debug = debug,
        )

        chunk_ids: typing.List[ int ] = []

        for chunk_id in chunk_iter:
            if chunk_id not in chunk_ids:
                chunk_ids.append(chunk_id)

        # enumerate chunks from a vector search -- the basic RAG process
        df_ann_chunk: pl.DataFrame = self.strw.chunk_table.search(  # type: ignore
            question
        ).limit(
            num_chunks
        ).to_polars()

        for row in df_ann_chunk.iter_rows(named = True):
            chunk_id: int =  row["uid"]  # type: ignore

            if chunk_id not in chunk_ids:
                chunk_ids.append(chunk_id)

        if debug:
            ic(chunk_ids)

        # get the text for the combined/ranked list of chunks
        id_list: str = ", ".join([ str(c_id) for c_id in chunk_ids ])
        filter_term: str = f"uid IN ({id_list})"

        chunks: typing.List[ str ] = self.strw.chunk_table.search().where(  # type: ignore
            filter_term
        ).select(
            [ "text" ]
        ).to_polars()["text"].to_list()

        return chunks[:num_chunks]


    def gen_subgraph_paths (
        self,
        anchor_nodes: set,
        *,
        debug: bool = False,
        ) -> typing.Iterator[ str ]:
        """
Generate pairwise shortest paths among the nodes from semantic
expansion, to define a subgraph.

In other words, this emulates a _semantic random walk_.
        """
        for pair in itertools.combinations(anchor_nodes, 2):
            if debug:
                ic(pair)

            for path in nx.all_shortest_paths(self.strw.sem_overlay, pair[0], pair[1]):
                if debug:
                    ic(path)

                for node in path:
                    if node not in pair:
                        dat: dict = self.strw.sem_overlay.nodes[node]

                        if debug:
                            ic(node, dat)

                        yield node


    def extract_subgraph (
        self,
        anchor_nodes: set,
        *,
        debug: bool = False,
        ) -> typing.Iterator[ int ]:
        """
Extract a subgraph, run a _centrality_ algorithm to rerank the most
referenced entities in the subgraph.
        """
        subgraph_iter: typing.Iterator[ str ] = self.gen_subgraph_paths(
            anchor_nodes,
            debug = debug,
        )

        subgraph: nx.Graph = self.strw.sem_overlay.subgraph(
            anchor_nodes.union(set(subgraph_iter))
        )

        rank_iter: dict = nx.pagerank(  # type: ignore
            subgraph,
            self.strw.config["tr"]["tr_alpha"],
        ).items()

        for node, rank in sorted(rank_iter, key = lambda x: x[1], reverse = True):
            dat: dict = self.strw.sem_overlay.nodes[node]

            if debug:
                ic(node, rank, dat)

            yield node


    def extract_chunk_neighbors (
        self,
        ranked_nodes: list,
        *,
        debug: bool = False,
        ) -> typing.Iterator[ int ]:
        """
Find the neighboring chunks for each _anchor node_ in the given list.
        """
        for node in ranked_nodes:
            if debug:
                ic(node)

            for neighbor in self.strw.sem_overlay.neighbors(node):
                dat: dict = self.strw.sem_overlay.nodes[neighbor]

                if dat["kind"] == "Chunk":
                    if debug:
                        ic(neighbor, dat)

                    chunk_id: int = int(neighbor.replace("chunk_", ""))
                    yield chunk_id


    def qa_cycle (
        self,
        question: str,
        *,
        debug: bool = False,
        ) -> baml_types.Response:
        """
Loop to answer questions.
        """
        chunks: typing.List[ str ] = self.get_chunks(
            question,
            debug = debug,
        )

        context: str = "\n".join( chunks )

        response: baml_types.Response = b.RAG(
            question,
            context,
        )

        return response



# ----------------------------------------------------------------------------
#
#  Welcome to Baml! To use this generated code, please run the following:
#
#  $ pip install baml
#
# ----------------------------------------------------------------------------

# This file was generated by BAML: please do not edit it. Instead, edit the
# BAML files and re-generate this code using: baml-cli generate
# baml-cli is available with the baml package.

import typing
import typing_extensions
from pydantic import BaseModel, ConfigDict

import baml_py

from . import types

StreamStateValueT = typing.TypeVar('StreamStateValueT')
class StreamState(BaseModel, typing.Generic[StreamStateValueT]):
    value: StreamStateValueT
    state: typing_extensions.Literal["Pending", "Incomplete", "Complete"]
# #########################################################################
# Generated classes (1)
# #########################################################################

class Response(BaseModel):
    question_type: typing.Optional[typing.Union[str, str, str]] = None
    question: typing.Optional[str] = None
    answer: typing.Optional[str] = None

# #########################################################################
# Generated type aliases (0)
# #########################################################################



# ----------------------------------------------------------------------------
#
#  Welcome to Baml! To use this generated code, please run the following:
#
#  $ pip install baml
#
# ----------------------------------------------------------------------------

# This file was generated by BAML: please do not edit it. Instead, edit the
# BAML files and re-generate this code using: baml-cli generate
# baml-cli is available with the baml package.

import typing
import typing_extensions

from . import stream_types, types
from .runtime import DoNotUseDirectlyCallManager, BamlCallOptions

class LlmResponseParser:
    __options: DoNotUseDirectlyCallManager

    def __init__(self, options: DoNotUseDirectlyCallManager):
        self.__options = options

    def RAG(
        self, llm_response: str, baml_options: BamlCallOptions = {},
    ) -> types.Response:
        result = self.__options.merge_options(baml_options).parse_response(function_name="RAG", llm_response=llm_response, mode="request")
        return typing.cast(types.Response, result)

    

class LlmStreamParser:
    __options: DoNotUseDirectlyCallManager

    def __init__(self, options: DoNotUseDirectlyCallManager):
        self.__options = options

    def RAG(
        self, llm_response: str, baml_options: BamlCallOptions = {},
    ) -> stream_types.Response:
        result = self.__options.merge_options(baml_options).parse_response(function_name="RAG", llm_response=llm_response, mode="stream")
        return typing.cast(stream_types.Response, result)

    


#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
An adapted TextRank algorithm implementation for NetworkX.
see copyright/license https://github.com/DerwenAI/strwythura/README.md
"""

from collections import defaultdict
import itertools
import typing

import networkx as nx
import numpy as np
import pandas as pd
import polars as pl

from .opt import calc_quantile_bins, stripe_column, root_mean_square
from .graph import Entity


def run_textrank (
    config: dict,
    lex_graph: nx.Graph,
    ) -> pl.DataFrame:
    """
Run eigenvalue centrality (i.e., _Personalized PageRank_) to rank the entities.
    """
    # build a dataframe of node ranks and counts
    df_rank: pd.DataFrame = pd.DataFrame.from_dict([  # type: ignore
        {
            "node_id": node,
            "weight": rank,
            "count": lex_graph.nodes[node]["count"],
        }
        for node, rank in nx.pagerank(
                lex_graph,
                alpha = config["tr"]["tr_alpha"],
                weight = "count",
        ).items()
    ])

    # normalize by column and calculate quantiles
    df1: pd.DataFrame = df_rank[[ "count", "weight" ]].apply(lambda x: x / x.max(), axis = 0)
    bins: np.ndarray = calc_quantile_bins(len(df1.index))

    # stripe each columns
    df2: pd.DataFrame = pd.DataFrame([
        stripe_column(values, bins)  # type: ignore
        for _, values in df1.items()
    ]).T

    # renormalize the ranks
    df_rank["rank"] = df2.apply(root_mean_square, axis = 1)
    rank_col: np.ndarray = df_rank["rank"].to_numpy()
    rank_col /= sum(rank_col)
    df_rank["rank"] = rank_col

    # move the ranked weights back into the graph
    for _, row in df_rank.iterrows():
        node: int = row["node_id"]
        lex_graph.nodes[node]["rank"] = row["rank"]

    df: pl.DataFrame = pl.DataFrame([
        node_attr
        for node, node_attr in lex_graph.nodes(data = True)
        if node_attr["kind"] == "Entity"
    ]).sort(
        [ "rank" ],
        descending = True,
    )

    return df


def cooccur_entities (
    lex_graph: nx.Graph,
    span_decoder: typing.Dict[ tuple, Entity ],
    ) -> None:
    """
Connect entities which co-occur within the same sentence.
    """
    ent_map: typing.Dict[ int, typing.Set[ int ]] = defaultdict(set)

    for ent in span_decoder.values():
        if ent.node is not None:
            ent_map[ent.sent_id].add(ent.node)

    for nodes in ent_map.values():
        for pair in itertools.combinations(list(nodes), 2):
            if not lex_graph.has_edge(*pair):
                lex_graph.add_edge(
                    pair[0],
                    pair[1],
                    rel = "CO_OCCURS_WITH",
                    prob = 1.0,
                )



# ----------------------------------------------------------------------------
#
#  Welcome to Baml! To use this generated code, please run the following:
#
#  $ pip install baml
#
# ----------------------------------------------------------------------------

# This file was generated by BAML: please do not edit it. Instead, edit the
# BAML files and re-generate this code using: baml-cli generate
# baml-cli is available with the baml package.

import typing
import typing_extensions
from enum import Enum


from pydantic import BaseModel, ConfigDict


import baml_py

CheckT = typing_extensions.TypeVar('CheckT')
CheckName = typing_extensions.TypeVar('CheckName', bound=str)

class Check(BaseModel):
    name: str
    expression: str
    status: str
class Checked(BaseModel, typing.Generic[CheckT, CheckName]):
    value: CheckT
    checks: typing.Dict[CheckName, Check]

def get_checks(checks: typing.Dict[CheckName, Check]) -> typing.List[Check]:
    return list(checks.values())

def all_succeeded(checks: typing.Dict[CheckName, Check]) -> bool:
    return all(check.status == "succeeded" for check in get_checks(checks))
# #########################################################################
# Generated enums (0)
# #########################################################################

# #########################################################################
# Generated classes (1)
# #########################################################################

class Response(BaseModel):
    question_type: typing.Union[typing_extensions.Literal['verbatim'], typing_extensions.Literal['reworded'], typing_extensions.Literal['new']]
    question: str
    answer: typing.Optional[str] = None

# #########################################################################
# Generated type aliases (0)
# #########################################################################



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Profiling utilities.
see copyright/license https://github.com/DerwenAI/strwythura/README.md
"""

import linecache
import tracemalloc
import typing

from pyinstrument import Profiler

KILO_B: float = 1024.0


class PerfProfiler:
    """
Profiling utilities for:

  * probablistic call trace
  * memory usage
    """

    def __init__ (
        self,
        ) -> None:
        """
Constructor.
        """
        self.profiler: Profiler = Profiler()


    def start (
        self,
        ) -> None:
        """
Start the profiling.
        """
        self.profiler.start()
        tracemalloc.start()


    def display_top (
        self,
        *,
        full: bool = False,
        key_type: str = "lineno",
        limit: int = 10,
        ) -> None:
        """
Display the top `limit` lines allocating the most memory,
ignoring `<frozen importlib._bootstrap>` and `<unknown>` files.
        """
        amount: tuple = tracemalloc.get_traced_memory()
        peak: float = round(amount[1] / KILO_B / KILO_B, 2)
        print(f"Peak memory usage: {peak} MB")

        if full:
            snapshot: tracemalloc.Snapshot = tracemalloc.take_snapshot().filter_traces((
                tracemalloc.Filter(False, "<frozen importlib._bootstrap>"),
                tracemalloc.Filter(False, "<unknown>"),
            ))

            print(f"Top {limit} lines for memory usage:")
            top_stats: list = snapshot.statistics(key_type)

            for index, stat in enumerate(top_stats[:limit], 1):
                frame: tracemalloc.Frame = stat.traceback[0]

                print(
                    "#%s: %s:%s: %.1f KiB"  # pylint: disable=C0209
                    % (index, frame.filename, frame.lineno, stat.size / KILO_B)
                )

                line: typing.Optional[ str ] = linecache.getline(
                    frame.filename,
                    frame.lineno,
                )

                if line is not None:
                    print(f"    {line.strip()}")

            other: typing.Optional[ list ] = top_stats[limit:]

            if other is not None:
                size: float = sum(stat.size for stat in other)
                print(f"{len(other)} other usage: {round(size / KILO_B, 1)} KiB")

            total: float = sum(stat.size for stat in top_stats)
            print(f"Total allocated size: {round(total / KILO_B, 1)} KiB")


    def report (
        self,
        ) -> None:
        """
Stop the call trace profiler from further sampling,
take a snapshot of the memory usage, then report
performance statistics.
        """
        # call trace
        self.profiler.stop()
        self.profiler.print()

        # memory usage
        self.display_top()
        tracemalloc.stop()
